
/*

 X-Power and Nordic Power Amiga 500 freezer cartridge ROM image
 decrypt/encrypt utility by Toni Wilen 2007-08-19
 
*/

#include <stdio.h>
#include <string.h>

#define SIZE 65536

static unsigned char bswap (unsigned char v,int b7,int b6,int b5,int b4,int b3,int b2,int b1,int b0)
{
    unsigned char b = 0;

    b |= ((v >> b7) & 1) << 7;
    b |= ((v >> b6) & 1) << 6;
    b |= ((v >> b5) & 1) << 5;
    b |= ((v >> b4) & 1) << 4;
    b |= ((v >> b3) & 1) << 3;
    b |= ((v >> b2) & 1) << 2;
    b |= ((v >> b1) & 1) << 1;
    b |= ((v >> b0) & 1) << 0;
    return b;
}
static unsigned char bswapx (unsigned char v,int b7,int b6,int b5,int b4,int b3,int b2,int b1,int b0)
{
    unsigned char b = 0;

    b |= ((v >> 7) & 1) << b7;
    b |= ((v >> 6) & 1) << b6;
    b |= ((v >> 5) & 1) << b5;
    b |= ((v >> 4) & 1) << b4;
    b |= ((v >> 3) & 1) << b3;
    b |= ((v >> 2) & 1) << b2;
    b |= ((v >> 1) & 1) << b1;
    b |= ((v >> 0) & 1) << b0;
    return b;
}
static unsigned short wswap (unsigned short v,int b15,int b14,int b13,int b12, int b11, int b10, int b9, int b8, int b7,int b6,int b5,int b4,int b3,int b2,int b1,int b0)
{
    unsigned short b = 0;

    b |= ((v >> b15) & 1) << 15;
    b |= ((v >> b14) & 1) << 14;
    b |= ((v >> b13) & 1) << 13;
    b |= ((v >> b12) & 1) << 12;
    b |= ((v >> b11) & 1) << 11;
    b |= ((v >> b10) & 1) << 10;
    b |= ((v >> b9) & 1) << 9;
    b |= ((v >> b8) & 1) << 8;
    b |= ((v >> b7) & 1) << 7;
    b |= ((v >> b6) & 1) << 6;
    b |= ((v >> b5) & 1) << 5;
    b |= ((v >> b4) & 1) << 4;
    b |= ((v >> b3) & 1) << 3;
    b |= ((v >> b2) & 1) << 2;
    b |= ((v >> b1) & 1) << 1;
    b |= ((v >> b0) & 1) << 0;
    return b;
}
static unsigned short wswapx (unsigned short v,int b15,int b14,int b13,int b12, int b11, int b10, int b9, int b8, int b7,int b6,int b5,int b4,int b3,int b2,int b1,int b0)
{
    unsigned short b = 0;

    b |= ((v >> 15) & 1) << b15;
    b |= ((v >> 14) & 1) << b14;
    b |= ((v >> 13) & 1) << b13;
    b |= ((v >> 12) & 1) << b12;
    b |= ((v >> 11) & 1) << b11;
    b |= ((v >> 10) & 1) << b10;
    b |= ((v >> 9) & 1) << b9;
    b |= ((v >> 8) & 1) << b8;
    b |= ((v >> 7) & 1) << b7;
    b |= ((v >> 6) & 1) << b6;
    b |= ((v >> 5) & 1) << b5;
    b |= ((v >> 4) & 1) << b4;
    b |= ((v >> 3) & 1) << b3;
    b |= ((v >> 2) & 1) << b2;
    b |= ((v >> 1) & 1) << b1;
    b |= ((v >> 0) & 1) << b0;
    return b;
}

#define AXOR 0x817f

// middle (even)
static void descramble1(unsigned char *buf, int size)
{
    int i;

    for (i = 0; i < size; i++)
	buf[i] = bswap(buf[i], 4, 1, 5, 3, 0, 7, 6, 2);
}
static void descramble1a(unsigned char *buf, int size)
{
    int i;
    unsigned char tbuf[SIZE];

    memcpy(tbuf, buf, size);
    for (i = 0; i < size; i++) {
	int a = (i ^ AXOR) & (size - 1);
	buf[i] = tbuf[wswap(a, 15, 9, 10, 4, 6, 5, 3, 8, 14, 13, 0, 12, 11, 2, 1, 7)];
    }
}

static void scramble1(unsigned char *buf, int size)
{
    int i;

    for (i = 0; i < size; i++)
	buf[i] = bswapx(buf[i], 4, 1, 5, 3, 0, 7, 6, 2);
}
static void scramble1a(unsigned char *buf, int size)
{
    int i, a;
    unsigned char tbuf[SIZE];

    a = AXOR & (size - 1);
    memcpy(tbuf, buf, size);
    for (i = 0; i < size; i++) {
	buf[i] = tbuf[a ^ wswapx(i, 15, 9, 10, 4, 6, 5, 3, 8, 14, 13, 0, 12, 11, 2, 1, 7)];
    }
}

// corner (odd)
static void descramble2(unsigned char *buf, int size)
{
    int i;

    for (i = 0; i < size; i++)
	buf[i] = bswap(buf[i], 5, 4, 3, 2, 1, 0, 7, 6);
}
static void descramble2a(unsigned char *buf, int size)
{
    int i;
    unsigned char tbuf[SIZE];

    memcpy(tbuf, buf, size);
    for (i = 0; i < size; i++) {
	int a = (i ^ AXOR) & (size - 1);
	buf[i] = tbuf[wswap(a, 15, 2, 4, 0, 1, 10, 11, 8, 13, 14, 12, 9, 7, 5, 6, 3)];
    }
}

static void scramble2(unsigned char *buf, int size)
{
    int i;

    for (i = 0; i < size; i++)
	buf[i] = bswapx(buf[i], 5, 4, 3, 2, 1, 0, 7, 6);
}
static void scramble2a(unsigned char *buf, int size)
{
    int i, a;
    unsigned char tbuf[SIZE];

    a = AXOR & (size - 1);
    memcpy(tbuf, buf, size);
    for (i = 0; i < size; i++) {
	buf[i] = tbuf[a ^wswapx(i, 15, 2, 4, 0, 1, 10, 11, 8, 13, 14, 12, 9, 7, 5, 6, 3)];
    }
}

int main (int argc, char *argv[])
{
    int i;
    FILE *fs;
    unsigned char buf[SIZE];
    unsigned char buf2[2 * SIZE];
    int size;
    int doscramble = -1;
    char c;

    memset(buf2, 0, 2 * SIZE);

    if (argc < 4) {
	printf("xnordicpowerdecrypt, de/encrypt X-Power and Nordic Power ROM images\n");
	printf(" decrypt: xnordicpowerdecrypt.exe d even.rom odd.rom newimage.rom\n");
	printf(" encrypt: xnordicpowerdecrypt.exe c image.rom neweven.rom newodd.rom\n");
	return 0;
    }

    c = toupper(argv[1][0]);
    if (c == 'D')
	doscramble = 0;
    if (c == 'S')
	doscramble = 1;
    if (doscramble < 0)
	return 0;

    if (doscramble) {

	fs = fopen(argv[2], "rb");
	if (!fs)
	    return 0;
	size = fread(buf2, 1, 2 * SIZE, fs);
	fclose(fs);
	size /= 2;
	for (i = 0; i < size; i++)
	    buf[i] = buf2[i * 2 + 0];
	scramble1a(buf, size);
	scramble1(buf, size);
	fs = fopen(argv[3], "wb");
	if (!fs)
	    return 0;
	fwrite(buf, 1, size, fs);
	fclose(fs);
	for (i = 0; i < size; i++)
	    buf[i] = buf2[i * 2 + 1];
	scramble2a(buf, size);
	scramble2(buf, size);
	fs = fopen(argv[4], "wb");
	if (!fs)
	    return 0;
	fwrite(buf, 1, size, fs);
	fclose(fs);


    } else {
	fs = fopen(argv[2], "rb");
	if (!fs)
	    return 0;
	size = fread(buf, 1, SIZE, fs);
	fclose(fs);
	descramble1(buf, size);
	descramble1a(buf, size);
	for (i = 0; i < size; i++)
	    buf2[i * 2 + 0] = buf[i];

	fs = fopen(argv[3], "rb");
	if (!fs)
	    return 0;
	fread(buf, 1, size, fs);
	fclose(fs);
	descramble2(buf, size);
	descramble2a(buf, size);
	for (i = 0; i < size; i++)
	    buf2[i * 2 + 1] = buf[i];

	fs = fopen(argv[4], "wb");
	if (!fs)
	    return 0;
	fwrite(buf2, 1, 2 * size, fs);
	fclose(fs);
    }

    return 0;
}